__all__ = [
    "co2tab",
]


def co2tab(argv=None):
    import os
    import shutil

    import pkg_resources

    parser = _get_parser()
    args = parser.parse_args(argv)

    # Check that target directory exists
    if not os.path.isdir(args.path):
        raise ValueError("Directory '{}' not found.".format(args.path))

    # Copy CO2TAB file
    filename = pkg_resources.resource_filename("toughio", "data/CO2TAB")
    output_filename = os.path.join(args.path, "CO2TAB")
    shutil.copy(filename, output_filename)


def _get_parser():
    import argparse

    # Initialize parser
    parser = argparse.ArgumentParser(
        description=("Write CO2TAB in target directory."),
        formatter_class=argparse.RawTextHelpFormatter,
    )

    # Output path
    parser.add_argument(
        "path", type=str, help="directory path",
    )

    return parser



blocks = {
    "TITLE",
    "ROCKS",
    "RPCAP",
    "FLAC",
    "CHEMP",
    "NCGAS",
    "MULTI",
    "SOLVR",
    "START",
    "PARAM",
    "SELEC",
    "INDOM",
    "MOMOP",
    "TIMES",
    "FOFT",
    "COFT",
    "GOFT",
    "GENER",
    "DIFFU",
    "OUTPU",
    "ELEME",
    "COORD",
    "CONNE",
    "INCON",
    "MESHM",
    "NOVER",
    "ENDCY",
}

_Parameters = {
    "title": "",
    "eos": "",
    "n_component": None,
    "n_phase": None,
    "n_component_incon": None,
    "flac": {},
    "chemical_properties": {},
    "non_condensible_gas": [],
    "isothermal": False,
    "start": True,
    "nover": False,
    "rocks": {},
    "rocks_order": None,
    "options": {},
    "extra_options": {},
    "more_options": {},
    "selections": {},
    "solver": {},
    "generators": [],
    "diffusion": [],
    "times": [],
    "element_history": [],
    "connection_history": [],
    "generator_history": [],
    "output": {},
    "elements": {},
    "elements_order": None,
    "coordinates": False,
    "connections": {},
    "connections_order": None,
    "initial_conditions": {},
    "initial_conditions_order": None,
    "meshmaker": {},
    "default": {},
}

Parameters = dict(_Parameters)

flac = {
    "creep": False,
    "porosity_model": None,
    "version": None,
}

chemical_properties = {
    "temperature_crit": None,
    "pressure_crit": None,
    "compressibility_crit": None,
    "pitzer_factor": None,
    "dipole_moment": None,
    "boiling_point": None,
    "vapor_pressure_a": None,
    "vapor_pressure_b": None,
    "vapor_pressure_c": None,
    "vapor_pressure_d": None,
    "molecular_weight": None,
    "heat_capacity_a": None,
    "heat_capacity_b": None,
    "heat_capacity_c": None,
    "heat_capacity_d": None,
    "napl_density_ref": None,
    "napl_temperature_ref": None,
    "gas_diffusivity_ref": None,
    "gas_temperature_ref": None,
    "exponent": None,
    "napl_viscosity_a": None,
    "napl_viscosity_b": None,
    "napl_viscosity_c": None,
    "napl_viscosity_d": None,
    "volume_crit": None,
    "solubility_a": None,
    "solubility_b": None,
    "solubility_c": None,
    "solubility_d": None,
    "oc_coeff": None,
    "oc_fraction": None,
    "oc_decay": None,
}

selections = {
    "integers": {k + 1: None for k in range(16)},
    "floats": None,
}

options = {
    "n_iteration": None,
    "n_cycle": None,
    "n_second": None,
    "n_cycle_print": None,
    "verbosity": None,
    "temperature_dependence_gas": None,
    "effective_strength_vapor": None,
    "t_ini": None,
    "t_max": None,
    "t_steps": None,
    "t_step_max": None,
    "t_reduce_factor": None,
    "gravity": None,
    "mesh_scale_factor": None,
    "eps1": None,
    "eps2": None,
    "w_upstream": None,
    "w_newton": None,
    "derivative_factor": None,
}

extra_options = {k + 1: None for k in range(24)}

more_options = {k + 1: None for k in range(40)}

solver = {
    "id": 3,
    "z_precond": "Z0",
    "o_precond": "O0",
    "rel_iter_max": 0.1,
    "eps": 1.0e-6,
}

generators = {
    "label": None,
    "name": None,
    "nseq": None,
    "nadd": None,
    "nads": None,
    "type": None,
    "times": None,
    "rates": None,
    "specific_enthalpy": None,
    "layer_thickness": None,
}

output = {
    "format": None,
    "variables": None,
}

elements = {
    "nseq": None,
    "nadd": None,
    "material": "",
    "volume": None,
    "heat_exchange_area": None,
    "permeability_modifier": None,
    "center": None,
}

connections = {
    "nseq": None,
    "nadd": None,
    "permeability_direction": None,
    "nodal_distances": None,
    "interface_area": None,
    "gravity_cosine_angle": None,
    "radiant_emittance_factor": None,
}

initial_conditions = {
    "porosity": None,
    "userx": [None for _ in range(5)],
    "values": [None for _ in range(4)],
}

meshmaker = {
    "type": None,
    "parameters": [],
    "angle": None,
}

xyz = {
    "type": None,
    "n_increment": None,
    "sizes": None,
}

rz2d = {
    "type": None,
    "radii": [],
    "n_increment": None,
    "size": None,
    "radius": None,
    "radius_ref": None,
    "thicknesses": [],
}

default = {
    "density": None,
    "porosity": None,
    "permeability": None,
    "conductivity": None,
    "specific_heat": None,
    "compressibility": None,
    "expansivity": None,
    "conductivity_dry": None,
    "tortuosity": None,
    "klinkenberg_parameter": None,
    "distribution_coefficient_3": None,
    "distribution_coefficient_4": None,
    "initial_condition": [None for _ in range(6)],
    "relative_permeability": {"id": None, "parameters": []},
    "capillarity": {"id": None, "parameters": []},
    "permeability_model": {"id": 1, "parameters": []},
    "equivalent_pore_pressure": {"id": 3, "parameters": []},
    "phase_composition": None,
}

eos = {
    "eos1": [1, 2, 2, 6],
    "eos2": [2, 3, 2, 6],
    "eos3": [2, 3, 2, 6],
    "eos4": [2, 3, 2, 6],
    "eos5": [2, 3, 2, 6],
    "eos7": [3, 3, 2, 6],
    "eos8": [3, 3, 3, 6],
    "eos9": [1, 1, 1, 6],
    "ewasg": [3, 4, 3, 6],
    "eco2n": [3, 4, 3, 6],
    "eco2n_v2": [3, 4, 3, 6],
    "eco2m": [3, 4, 4, 6],
    "tmvoc": None,
}

header = "----1----*----2----*----3----*----4----*----5----*----6----*----7----*----8"



from ._base import BaseRelativePermeability

__all__ = [
    "Linear",
]


class Linear(BaseRelativePermeability):

    _id = 1
    _name = "Linear"

    def __init__(self, slmin, sgmin, slmax, sgmax):
        """
        Linear function.

        Parameters
        ----------
        slmin : scalar
            Lower liquid saturation threshold (RP(1)).
        sgmin : scalar
            Lower gas saturation threshold (RP(2)).
        slmax : scalar
            Upper liquid saturation threshold (RP(3)).
        sgmax : scalar
            Upper gas saturation threshold (RP(4)).

        """
        if slmin >= slmax:
            raise ValueError()
        if sgmin >= sgmax:
            raise ValueError()
        self.parameters = [slmin, sgmin, slmax, sgmax]

    def _eval(self, sl, *args):
        """Linear function."""
        slmin, sgmin, slmax, sgmax = args
        kl = (
            1.0
            if sl >= slmax
            else 0.0
            if sl <= slmin
            else (sl - slmin) / (slmax - slmin)
        )

        sg = 1.0 - sl
        kg = (
            1.0
            if sg >= sgmax
            else 0.0
            if sg <= sgmin
            else (sg - sgmin) / (sgmax - sgmin)
        )

        return kl, kg

    @property
    def parameters(self):
        """Return model parameters."""
        return [self._slmin, self._sgmin, self._slmax, self._sgmax]

    @parameters.setter
    def parameters(self, value):
        if len(value) != 4:
            raise ValueError()
        self._slmin = value[0]
        self._sgmin = value[1]
        self._slmax = value[2]
        self._sgmax = value[3]



from __future__ import division, unicode_literals, with_statement

import logging

import numpy as np

from ...__about__ import __version__ as version
from .._helpers import _materials, get_material_key
from .._mesh import CellBlock, Mesh

__all__ = [
    "read",
    "write",
]


meshio_to_avsucd_type = {
    "vertex": "pt",
    "line": "line",
    "triangle": "tri",
    "quad": "quad",
    "tetra": "tet",
    "pyramid": "pyr",
    "wedge": "prism",
    "hexahedron": "hex",
}
avsucd_to_meshio_type = {v: k for k, v in meshio_to_avsucd_type.items()}


meshio_to_avsucd_order = {
    "vertex": [0],
    "line": [0, 1],
    "triangle": [0, 1, 2],
    "quad": [0, 1, 2, 3],
    "tetra": [0, 1, 3, 2],
    "pyramid": [4, 0, 1, 2, 3],
    "wedge": [3, 4, 5, 0, 1, 2],
    "hexahedron": [4, 5, 6, 7, 0, 1, 2, 3],
}


avsucd_to_meshio_order = {
    k: (v if k != "pyramid" else [1, 2, 3, 4, 0])
    for k, v in meshio_to_avsucd_order.items()
}


def read(filename):
    """Read AVS-UCD ASCII file."""
    with open(filename, "r") as f:
        out = read_buffer(f)
    return out


def read_buffer(f):
    # Skip comments and unpack first line
    while True:
        line = f.readline().strip()
        if not line.startswith("#"):
            break
    num_nodes, num_cells, num_node_data, num_cell_data = [
        int(i) for i in line.split()[:4]
    ]

    # Read nodes
    point_ids, points = _read_nodes(f, num_nodes)

    # Read cells
    cell_ids, cells, cell_data = _read_cells(f, num_cells, point_ids)

    # Read node data
    if num_node_data:
        point_data = _read_data(f, num_nodes, point_ids)
    else:
        point_data = {}

    # Read cell data
    if num_cell_data:
        cdata = _read_data(f, num_cells, cell_ids)
        sections = np.cumsum([len(c[1]) for c in cells[:-1]])
        for k, v in cdata.items():
            cell_data[k] = np.split(v, sections)

    return Mesh(points, cells, point_data=point_data, cell_data=cell_data)


def _read_nodes(f, num_nodes):
    data = []
    for _ in range(num_nodes):
        line = f.readline().strip()
        data += [[float(x) for x in line.split()]]
    data = np.array(data)
    points_ids = {int(pid): i for i, pid in enumerate(data[:, 0])}
    return points_ids, data[:, 1:]


def _read_cells(f, num_cells, point_ids):
    cells = []
    cell_ids = {}
    cell_data = {"avsucd:material": []}
    count = 0
    for _ in range(num_cells):
        line = f.readline().strip().split()
        cell_id = int(line[0])
        cell_mat = int(line[1])
        cell_type = avsucd_to_meshio_type[line[2]]
        corner = [point_ids[int(pid)] for pid in line[3:]]

        if len(cells) > 0 and cells[-1].type == cell_type:
            cells[-1].data.append(corner)
            cell_data["avsucd:material"][-1].append(cell_mat)
        else:
            cells.append(CellBlock(cell_type, [corner]))
            cell_data["avsucd:material"].append([cell_mat])

        cell_ids[cell_id] = count
        count += 1

    # Convert to numpy arrays
    for k, c in enumerate(cells):
        cells[k] = CellBlock(
            c.type, np.array(c.data)[:, avsucd_to_meshio_order[c.type]]
        )
        cell_data["avsucd:material"][k] = np.array(cell_data["avsucd:material"][k])
    return cell_ids, cells, cell_data


def _read_data(f, num_entities, entity_ids):
    line = f.readline().strip().split()
    data_size = [int(i) for i in line[1:]]

    labels = {}
    data = {}
    for i, dsize in enumerate(data_size):
        line = f.readline().strip().split(",")
        labels[i] = line[0].strip().replace(" ", "_")
        data[labels[i]] = (
            np.empty(num_entities) if dsize == 1 else np.empty((num_entities, dsize))
        )

    for _ in range(num_entities):
        line = f.readline().strip().split()
        eid = entity_ids[int(line[0])]
        j = 0
        for i, dsize in enumerate(data_size):
            if dsize == 1:
                data[labels[i]][eid] = float(line[j + 1])
            else:
                data[labels[i]][eid] = [
                    float(val) for val in line[j + 1 : j + 1 + dsize]
                ]
            j += dsize
    return data


def write(filename, mesh):
    """Write AVS-UCD ASCII file."""
    if mesh.points.shape[1] == 2:
        logging.warning(
            "AVS-UCD requires 3D points, but 2D points given. "
            "Appending 0 third component."
        )
        mesh.points = np.column_stack(
            [mesh.points[:, 0], mesh.points[:, 1], np.zeros(mesh.points.shape[0])]
        )

    with open(filename, "w") as f:
        # Write meshio version
        f.write("# Written by toughio v{}\n".format(version))

        # Write first line
        num_nodes = len(mesh.points)
        num_cells = sum(len(c.data) for c in mesh.cells)
        num_node_data = [
            1 if v.ndim == 1 else v.shape[1] for v in mesh.point_data.values()
        ]
        num_cell_data = [
            1 if np.concatenate(v).ndim == 1 else np.concatenate(v).shape[1]
            for k, v in mesh.cell_data.items()
            if k not in _materials
        ]
        num_node_data_sum = sum(num_node_data)
        num_cell_data_sum = sum(num_cell_data)
        f.write(
            "{} {} {} {} 0\n".format(
                num_nodes, num_cells, num_node_data_sum, num_cell_data_sum
            )
        )

        # Write nodes
        _write_nodes(f, mesh.points)

        # Write cells
        _write_cells(f, mesh.cells, mesh.cell_data, num_cells)

        # Write node data
        if num_node_data_sum:
            labels = mesh.point_data.keys()
            data_array = np.column_stack([v for v in mesh.point_data.values()])
            _write_data(
                f, labels, data_array, num_nodes, num_node_data, num_node_data_sum
            )

        # Write cell data
        if num_cell_data_sum:
            labels = [k for k in mesh.cell_data.keys() if k not in _materials]
            data_array = np.column_stack(
                [
                    np.concatenate(v)
                    for k, v in mesh.cell_data.items()
                    if k not in _materials
                ]
            )
            _write_data(
                f, labels, data_array, num_cells, num_cell_data, num_cell_data_sum
            )


def _write_nodes(f, points):
    for i, (x, y, z) in enumerate(points):
        f.write("{} {} {} {}\n".format(i + 1, x, y, z))


def _write_cells(f, cells, cell_data, num_cells):
    # Interoperability with other formats
    mat_data = get_material_key(cell_data)

    # Material array
    if mat_data:
        material = np.concatenate(cell_data[mat_data])
    else:
        material = np.zeros(num_cells, dtype=int)

    # Loop over cells
    i = 0
    for k, v in cells:
        for cell in v[:, meshio_to_avsucd_order[k]]:
            cell_str = " ".join(str(c + 1) for c in cell)
            f.write(
                "{} {} {} {}\n".format(
                    i + 1, int(material[i]), meshio_to_avsucd_type[k], cell_str
                )
            )
            i += 1


def _write_data(f, labels, data_array, num_entities, num_data, num_data_sum):
    num_data_str = " ".join(str(i) for i in num_data)
    f.write("{} {}\n".format(len(num_data), num_data_str))

    for label in labels:
        f.write("{}, real\n".format(label))

    data_array = np.column_stack((np.arange(1, num_entities + 1), data_array))
    np.savetxt(f, data_array, delimiter=" ", fmt=["%d"] + ["%.14e"] * num_data_sum)



from . import _cli, capillarity, meshmaker, relative_permeability
from .__about__ import (
    __author__,
    __author_email__,
    __license__,
    __version__,
    __website__,
)
from ._io import (
    Output,
    read_history,
    read_input,
    read_output,
    register_input,
    register_output,
    write_input,
    write_output,
)
from ._mesh import CellBlock, Mesh, from_meshio, from_pyvista
from ._mesh import read as read_mesh
from ._mesh import read_time_series
from ._mesh import register as register_mesh
from ._mesh import write as write_mesh
from ._mesh import write_time_series

__all__ = [
    "Mesh",
    "CellBlock",
    "Output",
    "meshmaker",
    "register_input",
    "register_output",
    "register_mesh",
    "read_history",
    "read_input",
    "read_output",
    "write_input",
    "write_output",
    "from_meshio",
    "from_pyvista",
    "read_mesh",
    "write_mesh",
    "read_time_series",
    "write_time_series",
    "relative_permeability",
    "capillarity",
    "_cli",
    "__version__",
    "__author__",
    "__author_email__",
    "__website__",
    "__license__",
]



from ._linear import Linear
from ._milly import Milly
from ._pickens import Pickens
from ._trust import TRUST
from ._van_genuchten import vanGenuchten

__all__ = [
    "Linear",
    "Pickens",
    "TRUST",
    "Milly",
    "vanGenuchten",
]



import meshio

from . import avsucd, flac3d, pickle, tecplot, tough
from ._helpers import read, read_time_series, register, write, write_time_series
from ._mesh import CellBlock, Mesh, from_meshio, from_pyvista

__all__ = [
    "register",
    "read",
    "write",
    "read_time_series",
    "write_time_series",
    "Mesh",
    "CellBlock",
    "from_meshio",
    "from_pyvista",
]



import copy

import helpers
import pytest

import toughio

write_read = lambda mesh, writer_kws, reader_kws: helpers.write_read(
    "hybrid_mesh",
    mesh,
    toughio.write_mesh,
    toughio.read_mesh,
    writer_kws=writer_kws,
    reader_kws=reader_kws,
)


@pytest.mark.parametrize(
    "mesh_ref, file_format, binary",
    [
        (helpers.tet_mesh, "avsucd", None),
        (helpers.tet_mesh, "flac3d", False),
        (helpers.tet_mesh, "flac3d", True),
        (helpers.tet_mesh, "pickle", None),
        (helpers.tet_mesh, "tecplot", None),
        (helpers.hex_mesh, "avsucd", None),
        (helpers.hex_mesh, "flac3d", False),
        (helpers.hex_mesh, "flac3d", True),
        (helpers.hex_mesh, "pickle", None),
        (helpers.hex_mesh, "tecplot", None),
        (helpers.hybrid_mesh, "avsucd", None),
        (helpers.hybrid_mesh, "flac3d", False),
        (helpers.hybrid_mesh, "flac3d", True),
        (helpers.hybrid_mesh, "pickle", None),
    ],
)
def test_meshio(mesh_ref, file_format, binary):
    mesh_ref = copy.deepcopy(mesh_ref)
    writer_kws = {"file_format": file_format}
    if binary is not None:
        writer_kws["binary"] = binary

    mesh = write_read(
        mesh=mesh_ref, writer_kws=writer_kws, reader_kws={"file_format": file_format},
    )
    if file_format == "flac3d":
        mesh_ref.cell_data.pop("c", None)

    helpers.allclose_mesh(mesh_ref, mesh)



from ._cylindric_grid import cylindric_grid
from ._helpers import from_meshmaker
from ._structured_grid import structured_grid
from ._triangulate import triangulate
from ._voxelize import voxelize

__all__ = [
    "cylindric_grid",
    "voxelize",
    "structured_grid",
    "triangulate",
    "from_meshmaker",
]



from ..._common import block_to_format, str2format
from ..._io.input.tough._helpers import write_record


def block(keyword):
    """Decorate block writing functions."""

    def decorator(func):
        from functools import wraps

        header = "----1----*----2----*----3----*----4----*----5----*----6----*----7----*----8"

        @wraps(func)
        def wrapper(f, *args):
            f.write("{}{}\n".format(keyword, header))
            func(f, *args)
            f.write("\n")

        return wrapper

    return decorator


def _write_eleme(labels, materials, volumes, nodes, material_name=None):
    """Return a generator that iterates over the records of block ELEME."""
    label_length = len(labels[0])
    fmt = block_to_format["ELEME"][label_length]
    fmt = str2format(fmt)

    iterables = zip(labels, materials, volumes, nodes)
    for label, material, volume, node in iterables:
        mat = (
            material_name[material]
            if material_name and material in material_name.keys()
            else material
        )
        mat = mat if isinstance(mat, str) else "{:>5}".format(str(mat))
        record = write_record(
            [
                label,  # ID
                None,  # NSEQ
                None,  # NADD
                mat,  # MAT
                volume,  # VOLX
                None,  # AHTX
                None,  # PMX
                node[0],  # X
                node[1],  # Y
                node[2],  # Z
            ],
            fmt=fmt,
        )
        yield record[0]


def _write_coord(nodes):
    """Return a generator that iterates over the records of block COORD."""
    fmt = "{:20.13e}{:20.13e}{:20.13e}\n"

    for node in nodes:
        record = fmt.format(*node)
        yield record


def _write_conne(clabels, isot, d1, d2, areas, angles):
    """Return a generator that iterates over the records of block CONNE."""
    label_length = len(clabels[0][0])
    fmt = block_to_format["CONNE"][label_length]
    fmt = str2format(fmt)

    iterables = zip(clabels, isot, d1, d2, areas, angles)
    for label, isot, d1, d2, area, angle in iterables:
        record = write_record(
            [
                "".join(label),  # ID1-ID2
                None,  # NSEQ
                None,  # NAD1
                None,  # NAD2
                isot,  # ISOT
                d1,  # D1
                d2,  # D2
                area,  # AREAX
                angle,  # BETAX
                None,  # SIGX
            ],
            fmt=fmt,
        )
        yield record[0]


def _write_incon(
    labels, values, porosity=None, userx=None, phase_composition=None, eos=None
):
    """Return a generator that iterates over the records of block INCON."""
    porosity = porosity if porosity is not None else [None] * len(labels)
    userx = userx if userx is not None else [None] * len(labels)
    phase_composition = (
        phase_composition if phase_composition is not None else [None] * len(labels)
    )
    label_length = len(labels[0])
    fmt = block_to_format["INCON"]

    iterables = zip(labels, values, porosity, userx, phase_composition)
    for label, value, phi, usrx, indicat0 in iterables:
        cond1 = any(v > -1.0e-9 for v in value)
        cond2 = phi is not None
        cond3 = usrx is not None
        if cond1 or cond2 or cond3:
            # Record 1
            values = [label, "", ""]
            ignore_types = [1, 2]

            if phi is not None:
                values.append(phi)
            else:
                values.append("")
                ignore_types.append(3)

            if eos == "tmvoc":
                if indicat0 is not None:
                    values.append(indicat0)
                else:
                    values.append("")
                    ignore_types.append(4)

            else:
                if usrx is not None:
                    values += list(usrx)
                else:
                    values += 3 * [""]
                    ignore_types += [4, 5, 6]

            fmt1 = str2format(
                fmt[eos][label_length] if eos in fmt else fmt["default"][label_length],
                ignore_types=ignore_types,
            )
            fmt1 = "{}\n".format("".join(fmt1[: len(values)]))
            record = fmt1.format(*values)

            # Record 2
            n = min(4, len(value))
            values = []
            ignore_types = []
            for i, v in enumerate(value[:n]):
                if v > -1.0e9:
                    values.append(v)
                else:
                    values.append("")
                    ignore_types.append(i)

            fmt2 = str2format(fmt[0], ignore_types=ignore_types)
            fmt2 = "{}\n".format("".join(fmt2[: len(values)]))
            record += fmt2.format(*values)

            # Record 3 (EOS7R)
            if len(value) > 4:
                values = []
                ignore_types = []
                for i, v in enumerate(value[n:]):
                    if v > -1.0e9:
                        values.append(v)
                    else:
                        values.append("")
                        ignore_types.append(i)

                fmt2 = str2format(fmt[0], ignore_types=ignore_types)
                fmt2 = "{}\n".format("".join(fmt2[: len(values)]))
                record += fmt2.format(*values)

            yield record
        else:
            continue



from __future__ import with_statement

import json

import numpy as np

__all__ = [
    "read",
    "write",
]


def read(filename):
    """
    Import json TOUGH input file.

    Parameters
    ----------
    filename : str
        Input file name.

    Returns
    -------
    dict
        TOUGH input parameters.

    """

    def to_int(data):
        """Return dict with integer keys instead of strings."""
        return {int(k): data[k] for k in sorted(data.keys())}

    with open(filename, "r") as f:
        parameters = json.load(f)

    if "extra_options" in parameters.keys():
        parameters["extra_options"] = to_int(parameters["extra_options"])

    if "more_options" in parameters.keys():
        parameters["more_options"] = to_int(parameters["more_options"])

    if "selections" in parameters.keys():
        parameters["selections"]["integers"] = to_int(
            parameters["selections"]["integers"]
        )

    return parameters


def write(filename, parameters):
    """
    Export TOUGH parameters to json.

    Parameters
    ----------
    filename : str
        Output file name.
    parameters : dict
        Parameters to export.

    """
    from copy import deepcopy

    def jsonify(x):
        """JSON serialize data."""
        if isinstance(x, (np.int32, np.int64)):
            return int(x)
        elif isinstance(x, (list, tuple)):
            return [jsonify(xx) for xx in x]
        elif isinstance(x, np.ndarray):
            return x.tolist()
        elif isinstance(x, dict):
            return {k: jsonify(v) for k, v in x.items()}
        else:
            return x

    with open(filename, "w") as f:
        parameters = deepcopy(parameters)
        parameters = jsonify(parameters)
        json.dump(parameters, f, indent=4)



from ._base import BaseRelativePermeability

__all__ = [
    "Pickens",
]


class Pickens(BaseRelativePermeability):

    _id = 2
    _name = "Pickens"

    def __init__(self, x):
        """
        Gas perfect mobile function.

        Parameters
        ----------
        x : scalar
            RP(1).

        """
        self.parameters = [x]

    def _eval(self, sl, *args):
        """Gas perfect mobile function."""
        (x,) = args
        return sl ** x, 1.0

    @property
    def parameters(self):
        """Return model parameters."""
        return [self._x]

    @parameters.setter
    def parameters(self, value):
        if len(value) != 1:
            raise ValueError()
        self._x = value[0]



__all__ = [
    "merge",
]


def merge(argv=None):
    import os

    parser = _get_parser()
    args = parser.parse_args(argv)

    # Check that input, MESH and INCON files exist
    head = os.path.split(args.infile)[0]
    mesh_filename = head + ("/" if head else "") + "MESH"
    incon_filename = head + ("/" if head else "") + "INCON"

    if not os.path.isfile(args.infile):
        raise ValueError("File '{}' not found.".format(args.infile))
    if not os.path.isfile(mesh_filename):
        raise ValueError("MESH file not found.")
    incon_exists = os.path.isfile(incon_filename)

    # Buffer input file
    with open(args.infile, "r") as f:
        input_file = list(f)

    # Check that input file has at least blocks ROCKS, PARAM, ENDFI or ENDCY
    count = 0
    for line in input_file:
        count += int(line.upper()[:5] in {"ROCKS", "PARAM", "ENDFI", "ENDCY"})
    if count < 3:
        raise ValueError("Invalid input file '{}'.".format(args.infile))

    # Buffer MESH
    with open(mesh_filename, "r") as f:
        mesh_file = list(f)
    if not mesh_file[0].startswith("ELEME"):
        raise ValueError("Invalid MESH file.")

    # Buffer INCON if exist
    if incon_exists:
        with open(incon_filename, "r") as f:
            incon_file = list(f)
        if not incon_file[0].startswith("INCON"):
            raise ValueError("Invalid INCON file.")

    # Locate ENDFI or ENDCY
    for i, line in enumerate(input_file):
        if line.upper()[:5] in {"ENDFI", "ENDCY"}:
            break

    # Buffer output file
    output_file = input_file[:i]
    output_file += mesh_file
    if incon_exists:
        output_file += incon_file
    output_file += input_file[i:]

    # Write output file
    with open(args.outfile, "w") as f:
        for line in output_file:
            f.write(line)


def _get_parser():
    import argparse

    # Initialize parser
    parser = argparse.ArgumentParser(
        description=(
            "Merge input file, MESH and/or INCON into a single file. "
            "The files must be in the same directory."
        ),
        formatter_class=argparse.RawTextHelpFormatter,
    )

    # Input file
    parser.add_argument(
        "infile", type=str, help="TOUGH input file",
    )

    # Output file
    parser.add_argument("outfile", type=str, help="merged TOUGH input file")

    return parser



class FileIterator:
    def __init__(self, f, count=0):
        """File iterator helper class."""
        self.f = f
        self.count = count
        self.fiter = iter(f.readline, "")

    def __iter__(self):
        """Return iterator."""
        return self

    def __next__(self):
        """Return next item."""
        self.count += 1
        return next(self.fiter)

    def next(self):
        """Return next line."""
        return self.__next__()

    def seek(self, i, increment):
        """Set file's position."""
        self.count += increment
        self.f.seek(i)

    def tell(self):
        """Return current position of file."""
        return self.f.tell()



from __future__ import print_function

import numpy as np

__all__ = [
    "export",
]


format_to_ext = {
    "tecplot": ".dat",
    "vtk": ".vtk",
    "vtu": ".vtu",
    "xdmf": ".xdmf",
}


def export(argv=None):
    import os
    import sys

    from .. import read_mesh, read_output, write_time_series
    from .._io.output._common import reorder_labels
    from ..meshmaker import triangulate, voxelize

    parser = _get_parser()
    args = parser.parse_args(argv)

    # Check that TOUGH output and mesh file exist
    if not os.path.isfile(args.infile):
        raise ValueError("TOUGH output file '{}' not found.".format(args.infile))

    with_mesh = bool(args.mesh)
    if with_mesh:
        if not os.path.isfile(args.mesh):
            raise ValueError("Mesh file '{}' not found.".format(args.mesh))
        if args.voxelize:
            print("Mesh file has been provided. Skipping option --voxelize.")
    else:
        if args.voxelize:
            if not args.origin:
                raise ValueError("Option --voxelize requires option --origin.")
            elif len(args.origin) != 3:
                raise ValueError(
                    "Option --origin requires 3 parameters, got {}.".format(
                        len(args.origin)
                    )
                )

    # Read output file
    print("Reading file '{}' ...".format(args.infile), end="")
    sys.stdout.flush()
    output = read_output(args.infile)
    if args.file_format != "xdmf":
        if args.time_step is not None:
            if not (-len(output) <= args.time_step < len(output)):
                raise ValueError("Inconsistent time step value.")
            output = output[args.time_step]
        else:
            output = output[-1]
        input_format = output.format
    else:
        input_format = output[-1].format
    print(" Done!")

    # Display warning if mesh is provided but input file format is 'tecplot'
    # Continue as if mesh was not provided
    if input_format == "tecplot":
        with_mesh = False
        msg = (
            "Cannot use mesh file with Tecplot TOUGH output, inferring dimensionality"
            if args.mesh
            else "Inferring dimensionality"
        )
    else:
        with_mesh = bool(args.mesh)
        msg = "Mesh file not specified, inferring dimensionality"

    # Triangulate or voxelize if no mesh
    voxelized = False
    if not with_mesh:
        print("{} ...".format(msg), end="")
        sys.stdout.flush()
        points, axis = _get_points(output if args.file_format != "xdmf" else output[0])
        ndim = len(axis)
        print(" Done!")

        if args.voxelize or ndim == 1:
            if ndim == 1:
                if not args.origin:
                    raise ValueError(
                        "Mesh is {}D and requires option --origin.".format(ndim)
                    )
                elif len(args.origin) != 3:
                    raise ValueError(
                        "Option --origin requires 3 parameters, got {}.".format(
                            len(args.origin)
                        )
                    )

            print("Mesh is {}D, voxelizing mesh ...".format(ndim), end="")
            sys.stdout.flush()

            mesh = voxelize(points, args.origin, layer=args.layer)
            mesh.cell_dada = {}

            idx = np.arange(len(points))
            idx = np.array(
                [
                    x
                    for x, _ in sorted(
                        zip(idx, points), key=lambda x: (x[1][2], x[1][1], x[1][0])
                    )
                ]
            )

            if args.file_format != "xdmf":
                for label, data in output.data.items():
                    if label not in {"X", "Y", "Z"}:
                        mesh.add_cell_data(label, data[idx])

            voxelized = True

        else:
            print(
                "Mesh is {}D, performing point triangulation ...".format(ndim), end=""
            )
            sys.stdout.flush()

            mesh = triangulate(points[:, axis])
            mesh.cell_dada = {}

            if args.file_format != "xdmf":
                for label, data in output.data.items():
                    if label not in {"X", "Y", "Z"}:
                        mesh.add_point_data(label, data)

    else:
        print("Reading mesh file '{}' ...".format(args.mesh), end="")
        sys.stdout.flush()

        try:
            mesh = read_mesh(args.mesh)
        except Exception as e:
            raise ValueError("Unable to read mesh file {}: {}.".format(args.mesh, e))

        if args.file_format != "xdmf":
            mesh.point_data = {}
            mesh.cell_dada = {}
            mesh.field_data = {}
            mesh.point_sets = {}
            mesh.cell_sets = {}
            mesh.read_output(output)
        else:
            output = [reorder_labels(data, mesh.labels) for data in output]
    print(" Done!")

    # Output file name
    if args.output_file:
        filename = args.output_file
    else:
        head, _ = os.path.splitext(args.infile)
        filename = "{}{}".format(head, format_to_ext[args.file_format])

    # Write output file
    print("Writing output file '{}' ...".format(filename), end="")
    sys.stdout.flush()

    if args.file_format != "xdmf":
        mesh.write(filename, file_format=args.file_format)

    else:
        if voxelized:
            data = [
                {k: v[idx] for k, v in out.data.items() if k not in {"X", "Y", "Z"}}
                for out in output
            ]
        else:
            data = [
                {k: v for k, v in out.data.items() if k not in {"X", "Y", "Z"}}
                for out in output
            ]
        time_steps = [out.time for out in output]

        if with_mesh or voxelized:
            write_time_series(
                filename,
                mesh.points,
                mesh.cells,
                cell_data=data,
                time_steps=time_steps,
            )
        else:
            write_time_series(
                filename,
                mesh.points,
                mesh.cells,
                point_data=data,
                time_steps=time_steps,
            )

    print(" Done!")


def _get_parser():
    import argparse

    # Initialize parser
    parser = argparse.ArgumentParser(
        description=("Export TOUGH simulation results to a file for visualization."),
        formatter_class=argparse.RawTextHelpFormatter,
    )

    # Input file
    parser.add_argument(
        "infile", type=str, help="TOUGH output file",
    )

    # Mesh file
    parser.add_argument(
        "--mesh", "-m", type=str, default=None, help="pickled toughio.Mesh",
    )

    # Time step
    parser.add_argument(
        "--time-step", "-t", type=int, default=None, help="time step to export",
    )

    # Output file
    parser.add_argument(
        "--output-file", "-o", type=str, default=None, help="exported file",
    )

    # File format
    parser.add_argument(
        "--file-format",
        "-f",
        type=str,
        choices=("tecplot", "vtk", "vtu", "xdmf"),
        default="vtk",
        help="exported file format",
    )

    # Voxelize
    parser.add_argument(
        "--voxelize",
        "-v",
        default=False,
        action="store_true",
        help="voxelize mesh (only if mesh is not provided, requires option --origin)",
    )

    # Origin
    parser.add_argument(
        "--origin",
        nargs="+",
        type=float,
        default=None,
        help="coordinates of origin point (only if option --voxelize is enabled)",
    )

    # Layer
    parser.add_argument(
        "--layer",
        default=False,
        action="store_true",
        help="voxelize mesh by layers (only if option --voxelize is enabled)",
    )

    return parser


def _get_points(output):
    # Number of data points
    n_points = len(next(iter(output.data.values())))

    # Make sure that point coordinates exist
    count = 0

    if "X" in output.data.keys():
        X = np.array(output.data["X"])
        count += 1
    else:
        X = np.zeros(n_points)

    if "Y" in output.data.keys():
        Y = np.array(output.data["Y"])
        count += 1
    else:
        Y = np.zeros(n_points)

    if "Z" in output.data.keys():
        Z = np.array(output.data["Z"])
        count += 1
    else:
        Z = np.zeros(n_points)

    if count == 0:
        raise ValueError("No coordinate array ('X', 'Y', 'Z') found.")

    # Assert dimension of the problem
    nx = len(np.unique(X))
    ny = len(np.unique(Y))
    nz = len(np.unique(Z))

    # Reconstruct points cloud
    zeros = np.zeros(n_points)
    if ny == 1 and nz == 1:
        points = np.column_stack((X, zeros, zeros))
        axis = [0]
    elif nx == 1 and nz == 1:
        points = np.column_stack((zeros, Y, zeros))
        axis = [1]
    elif nx == 1 and ny == 1:
        points = np.column_stack((zeros, zeros, Z))
        axis = [2]
    elif nx == 1:
        points = np.column_stack((zeros, Y, Z))
        axis = [1, 2]
    elif ny == 1:
        points = np.column_stack((X, zeros, Z))
        axis = [0, 2]
    elif nz == 1:
        points = np.column_stack((X, Y, zeros))
        axis = [0, 1]
    else:
        points = np.column_stack((X, Y, Z))
        axis = [0, 1, 2]

    return points, axis
